# ▄████▄    ██████ ▓█████▄  ▒█████   █     █░███▄    █   ██████  ▄████▄   ▄▄▄       ██▓    ▓█████ 
#▒██▀ ▀█  ▒██    ▒ ▒██▀ ██▌▒██▒  ██▒▓█░ █ ░█░██ ▀█   █ ▒██    ▒ ▒██▀ ▀█  ▒████▄    ▓██▒    ▓█   ▀ 
#▒▓█    ▄ ░ ▓██▄   ░██   █▌▒██░  ██▒▒█░ █ ░█▓██  ▀█ ██▒░ ▓██▄   ▒▓█    ▄ ▒██  ▀█▄  ▒██░    ▒███   
#▒▓▓▄ ▄██▒  ▒   ██▒░▓█▄   ▌▒██   ██░░█░ █ ░█▓██▒  ▐▌██▒  ▒   ██▒▒▓▓▄ ▄██▒░██▄▄▄▄██ ▒██░    ▒▓█  ▄ 
#▒ ▓███▀ ░▒██████▒▒░▒████▓ ░ ████▓▒░░░██▒██▓▒██░   ▓██░▒██████▒▒▒ ▓███▀ ░ ▓█   ▓██▒░██████▒░▒████▒
#░ ░▒ ▒  ░▒ ▒▓▒ ▒ ░ ▒▒▓  ▒ ░ ▒░▒░▒░ ░ ▓░▒ ▒ ░ ▒░   ▒ ▒ ▒ ▒▓▒ ▒ ░░ ░▒ ▒  ░ ▒▒   ▓▒█░░ ▒░▓  ░░░ ▒░ ░
#  ░  ▒   ░ ░▒  ░ ░ ░ ▒  ▒   ░ ▒ ▒░   ▒ ░ ░ ░ ░░   ░ ▒░░ ░▒  ░ ░  ░  ▒     ▒   ▒▒ ░░ ░ ▒  ░ ░ ░  ░
#░        ░  ░  ░   ░ ░  ░ ░ ░ ░ ▒    ░   ░    ░   ░ ░ ░  ░  ░  ░          ░   ▒     ░ ░      ░   
#░ ░            ░     ░        ░ ░      ░            ░       ░  ░ ░            ░  ░    ░  ░   ░  ░
#░                  ░                                           ░                                 
## TODO: Remove once CSDownscale is on CRAN
## TODO: Move recipe checks to check_recipe()
source('modules/Downscaling/tmp/Interpolation.R')
source('modules/Downscaling/tmp/Intbc.R')
source('modules/Downscaling/tmp/Intlr.R')
source('modules/Downscaling/tmp/Analogs.R')
source('modules/Downscaling/tmp/LogisticReg.R')
source('modules/Downscaling/tmp/Utils.R')
#source("https://earth.bsc.es/gitlab/external/cstools/-/raw/master/R/CST_BiasCorrection.R")


## Entry params data and recipe?
downscale_datasets <- function(recipe, data) {
  # Function that downscale the hindcast using the method stated in the
  # recipe. For the moment, forecast must be null.
  #
  # data: list of s2dv_cube objects containing the hcst, obs and fcst.
  # recipe: object obtained when passing the .yml recipe file to read_yaml()
  
  type <- tolower(recipe$Analysis$Workflow$Downscaling$type)

  if (!is.null(data$fcst)) {
    warn(recipe$Run$logger,
	 "The downscaling will be only performed to the hindcast data")
    data$fcst <- NULL
  }

  if (type == "none") {
    hcst_downscal <- data$hcst
    DOWNSCAL_MSG <- "##### NO DOWNSCALING PERFORMED #####"
  } else {
    # Downscaling function params
    int_method     <- tolower(recipe$Analysis$Workflow$Downscaling$int_method)
    bc_method      <- tolower(recipe$Analysis$Workflow$Downscaling$bc_method)
    cal_method     <- tolower(recipe$Analysis$Workflow$Downscaling$cal_method)
    lr_method      <- tolower(recipe$Analysis$Workflow$Downscaling$lr_method)
    logreg_method <- tolower(recipe$Analysis$Workflow$Downscaling$logreg_method)
    target_grid    <- tolower(recipe$Analysis$Workflow$Downscaling$target_grid)
    nanalogs       <- as.numeric(recipe$Analysis$Workflow$Downscaling$nanalogs)
      
 ## TODO: Compute number of cores
    if (is.null(recipe$Analysis$ncores)) {
      ncores <- 1
    } else {
      ncores <- recipe$Analysis$ncores
    }
    ## TODO: add the parametre loocv where it corresponds
    if (is.null(recipe$Analysis$loocv)) {
      loocv <- TRUE
    } else {
      loocv <- recipe$Analysis$loocv
    }
    # Define downscaling options
    DOWNSCAL_TYPES <- c("none", "int", "intbc", "intlr", "analogs", "logreg")
    BC_METHODS <- c("simple_bias", "calibration", "quantile_mapping")
    LR_METHODS <- c("basic", "large_scale", "4nn")
    LOGREG_METHODS <- c("ens_mean", "ens_mean_sd", "sorted_members")

    # Interpolation
    if (type == "int") {
      ## TODO: Move this check elsewhere
      # Ensure that observations are in the same grid as experiments
      # Only needed for this method because the others already return the 
      # observations
      latmin <- data$hcst$lat[1]
      lonmin <- data$hcst$lon[1]
      latmax <- data$hcst$lat[length(data$hcst$lat)]
      lonmax <- data$hcst$lon[length(data$hcst$lon)]
      # Interpolate hcst
      hcst_downscal <- CST_Interpolation(data$hcst, 
                                         points = NULL, 
                                         method_remap = int_method, 
                                         target_grid = target_grid,
                                         lat_dim = "latitude", 
                                         lon_dim = "longitude", 
                                         region = c(lonmin, lonmax, latmin, latmax),
                                         method_point_interp = NULL)
      # Interpolate obs
      obs_downscal <- CST_Interpolation(data$obs,
                                        points = NULL,
                                        method_remap = int_method,
                                        target_grid = target_grid,
                                        lat_dim = "latitude",
                                        lon_dim = "longitude",
                                        region = c(lonmin, lonmax, latmin, latmax),
                                        method_point_interp = NULL)
      hcst_downscal$obs <- obs_downscal$exp

    } else if (type == "intbc" & (bc_method=="cal" | bc_method=="calibration")) {
     # Interpolate hcst and obs with bias correction (via calibration)
      hcst_downscal <- CST_Intbc(data$hcst, data$obs, 
                                 target_grid = target_grid, 
                                 bc_method = bc_method, 
                                 int_method = int_method, 
                                 cal.method=cal_method,
                                 points = NULL,
                                 method_point_interp = NULL, 
                                 lat_dim = "latitude", 
                                 lon_dim = "longitude",
                                 sdate_dim = "syear", 
                                 member_dim = "ensemble", 
                                 region = NULL, 
                                 ncores = ncores)

      } else if (type == "intbc" & !(bc_method=="cal" | bc_method=="calibration")) {
      # Interpolate hcst and obs with bias correction (any other method than calibration)
      hcst_downscal <- CST_Intbc(data$hcst, data$obs, 
                                 target_grid = target_grid, 
                                 bc_method = bc_method, 
                                 int_method = int_method, 
                                 points = NULL,
                                 method_point_interp = NULL, 
                                 lat_dim = "latitude", 
                                 lon_dim = "longitude",
                                 sdate_dim = "syear", 
                                 member_dim = "ensemble", 
                                 region = NULL, 
                                 ncores = ncores)

      } else if (type == "intlr") {
      ## TODO: add the possibility to have the element 'pred' in 'data'
      if (lr_method == "large_scale") {
        if (is.null(data$pred$data)) {
          stop("Please provide the large scale predictors in the element 'data$pred$data'.")
        }
      } else {
        data$pred$data <- NULL
      }
      # Interpolate hcst and obs with linear regression
      hcst_downscal <- CST_Intlr(data$hcst, data$obs, 
                                 lr_method = lr_method, 
                                 target_grid = target_grid, 
                                 points = NULL, 
                                 int_method = int_method,
                                 method_point_interp = NULL, 
                                 predictors = data$pred$data, 
                                 lat_dim = "latitude",
                                 lon_dim = "longitude",
                                 sdate_dim = "syear",
                                 time_dim = "time",
                                 member_dim = "ensemble",
                                 large_scale_predictor_dimname = 'vars',
                                 loocv = loocv, 
                                 region = NULL,
                                 ncores = ncores)
    } else if (type == "analogs") {
      if (length(nanalogs) == 0) {
	info(recipe$Run$logger,
	     paste("The number of analogs for searching has not been provided",
		   "in the recipe. Setting it to 3."))
        nanalogs <- 3
      }
      # Apply analogs method to hcst and obs
      hcst_downscal <- CST_Analogs(data$hcst, data$obs, 
                                   grid_exp = data$hcst$source_files[1], 
                                   nanalogs = nanalogs, 
                                   fun_analog = "min", 
                                   lat_dim = "latitude", 
                                   lon_dim = "longitude",
                                   sdate_dim = "syear", 
                                   time_dim = "time", 
                                   member_dim = "ensemble",
                                   region = NULL, 
                                   return_indices = FALSE,
                                   loocv_window = loocv, 
                                   ncores = ncores)
    } else if (type == "logreg") {
      # Since we are forcing to create three categories, and applying cross-validation,
      # we need at least six years of data for the logistic regression function to not 
      # crash
      if (dim(data$hcst$data)[["syear"]] < 6) {
	error(recipe$Run$logger,
	      paste("The number of years of data is insufficient for the",
		    "logistic regression method. Please provide six or more."))
	stop()
      }
      # Apply logistic regression to hcst and obs
      hcst_downscal <- CST_LogisticReg(data$hcst, data$obs, 
                                       target_grid = target_grid, 
                                       int_method = int_method, 
                                       log_reg_method = log_reg_method,
                                       probs_cat = c(1/3,2/3), 
                                       return_most_likely_cat = FALSE, 
                                       points = NULL,
                                       method_point_interp = NULL, 
                                       lat_dim = "latitude", 
                                       lon_dim = "longitude", 
                                       sdate_dim = "syear",
                                       member_dim = "ensemble", 
                                       region = NULL, 
                                       loocv = loocv, 
                                       ncores = ncores)

    }
    DOWNSCAL_MSG <- "##### DOWNSCALING COMPLETE #####"
  } 
  info(recipe$Run$logger, DOWNSCAL_MSG)
  return(list(hcst = hcst_downscal$exp, obs = hcst_downscal$obs, fcst = NULL))
}   
 
